# Copyright (C) 2007, 2008 by Intevation GmbH
# Authors:
# Bernhard Herzog <bh@intevation.de>
#
# This program is free software under the GPL (>=v2)
# Read the file GPL.txt coming with the software for details.

"""Parse the OSAAS form data into a record"""

import cgi
import datetime
import time

class FormParserException(Exception):

    pass


class Field(object):

    def __init__(self, name, required=False):
        self.name = name
        self.required = required

    def deserialize(self, s):
        return s


class DateField(Field):

    def deserialize(self, string):
        return datetime.datetime(*time.strptime(string.strip(),
                                                "%Y-%m-%dT%H:%M:%SZ")[:6])



class Record:

    def __init__(self):
        self._fields = []

    def set_fields(self, fields, formdata, fieldname_normalizer=None):
        try:
            parsed_form = cgi.parse_qs(formdata, keep_blank_values=True,
                                       strict_parsing=True)
        except ValueError, err:
            raise FormParserException("Error parsing form data: %s" % err)

        if fieldname_normalizer is not None:
            for name, value in parsed_form.items():
                normalized = fieldname_normalizer(name)
                del parsed_form[name]
                parsed_form[normalized] = value

        for field in fields:
            value = parsed_form.get(field.name)
            if value is not None:
                if len(value) == 1:
                    setattr(self, field.name, field.deserialize(value[0]))
                else:
                    raise FormParserException("only one value allowed"
                                              " for field %r" % field.name)
            else:
                setattr(self, field.name, None)

        self._fields.extend(fields)

    def validate_fields(self):
        for field in self._fields:
            if field.required and getattr(self, field.name, None) is None:
                raise FormParserException("formdata is missing required"
                                          " field %r" % field.name)


form_fields = [Field(name, required=True)
               for name in ["user", "wmsidextern", "wmsidintern",
                            "requeststring"]]
form_fields.append(DateField("responsetime", required=True))

request_fields = [
    Field("VERSION", required=True),
    Field("REQUEST", required=True),
    Field("WIDTH", required=True),
    Field("HEIGHT", required=True),
    Field("BBOX", required=True),
    Field("LAYERS", required=True),
    Field("SERVICE"),
    Field("FORMAT"),
    Field("TRANSPARENT"),
    Field("EXCEPTIONS"),
    Field("BGCOLOR"),
    Field("STYLES"),
    Field("SRS"),
    ]

def parse_formdata(formdata):
    record = Record()
    record.set_fields(form_fields, formdata)
    record.set_fields(request_fields, record.requeststring,
                      fieldname_normalizer=lambda s: s.upper())
    record.raw_record = formdata
    return record
