# Copyright (C) 2007 by Intevation GmbH
# Authors:
# Bernhard Herzog <bh@intevation.de>
#
# This program is free software under the GPL (>=v2)
# Read the file GPL.txt coming with the software for details.

"""HTTP server that can run in a thread and be stopped"""

import socket
import errno
import threading
import BaseHTTPServer
import logging


class HTTPServer(BaseHTTPServer.HTTPServer):

    """HTTP server that can be run in a thread and be stopped"""

    do_shutdown = False

    def __init__(self, *args, **kw):
        for name, value in [("access_logger_name", "httpserver.access"),
                            ("error_logger_name", "httpserver.error")]:
            if kw.has_key(name):
                value = kw[name]
                del kw[name]
            setattr(self, name, value)
        BaseHTTPServer.HTTPServer.__init__(self, *args, **kw)

    def serve_forever(self):
        """Handle requests until self.do_shutdown is True."""
        while not self.do_shutdown:
            self.handle_request()

    def server_close(self):
        """Shutdown the server.

        Usually this is called from a thread while another thread
        executes the serve_forever method.  In that case the other
        thread is most likely blocked while listening for new
        connections.  After this method returns, the caller will have to
        do something to unblock the other thread, such as creating a TCP
        connection to the server.
        """
        BaseHTTPServer.HTTPServer.server_close(self)
        self.do_shutdown = True


class HTTPRequestHandler(BaseHTTPServer.BaseHTTPRequestHandler):

    def log_message(self, format, *args):
        """Log an arbitrary message.

        Overrides the base class version to write the message into the
        logger given by the server's access_logger_name.  The messge is
        logged with INFO level.
        """
        logger = logging.getLogger(self.server.access_logger_name)
        logger.info("%s - - [%s] %s",
                    self.address_string(),
                    self.log_date_time_string(),
                    format % args)

    def log_error(self, *args):
        """Log an error.

        Overrides the base class method so that the message is written
        to the logger given by the server's error_logger_name.  The
        message is logged with the ERROR level.
        """
        logger = logging.getLogger(self.server.error_logger_name)
        logger.error(*args)


class ServerThread:

    """Class to run a HTTPServer instance in a thread"""

    def __init__(self, server):
        """Initializes the server thread with an instance of HTTPServer"""
        self.server = server
        self.server_port = self.server.server_port
        self.server_thread = None

    def start(self, daemon=False):
        """Starts the server thread"""
        self.server_thread = threading.Thread(target=self._serve_forever)
        self.server_thread.setDaemon(daemon)
        self.server_thread.start()

    def _serve_forever(self):
        """Helper method to run the server's serve_forever method in a thread"""
        self.server.serve_forever()

    def stop(self):
        """Stops the server thread"""
        if self.server_thread is None:
            return

        self.server.server_close()

        # The server thread might be blocked while listening on the
        # socket.  Unblock it by connecting to the port.
        try:
            s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            s.connect(("localhost", self.server_port))
            s.close()
        except socket.error, exc:
            # The server may have shut down already when we try to
            # connect, so we ignore connection failures.
            if exc[0] == errno.ECONNREFUSED:
                pass
            else:
                raise

        self.server_thread.join()
        self.server_thread = None
