# Copyright (C) 2007 by Intevation GmbH
# Authors:
# Bernhard Herzog <bh@intevation.de>
#
# This program is free software under the GPL (>=v2)
# Read the file GPL.txt coming with the software for details.

"""Simple pool of worker threads to process tasks like e.g. HTTP requests"""


import traceback
import threading
import Queue


class ThreadPool(object):

    """A class to manage a set of threads that read tasks from a queue.

    The threads managed by the class are started as deamon threads so
    that it's not required to terminate them if the program using them
    wants to exit.  Still, they can be explicitly stopped with the stop
    method.

    The threads read tasks from a Queue object managed by the ThreadPool
    instance using the queue's blocking get() method.  Tasks taken from
    the queue are then passed to the handler_function which was
    specified then the ThreadPool instance was created.  The
    handler_function is called with one parameter, the task object
    (which can be any object except None), and should return when the
    task is finished.  Repeatedly getting new tasks from the queue is
    handled by the thread pool itself.

    The program using the thread pool puts tasks into the queue with the
    thread pool's put method.
    """

    def __init__(self, num_threads, handler_function, logger=None):
        """Initialize the thread pool with a number of threads and a handler"""
        self.queue = Queue.Queue(0)
        self.num_threads = num_threads
        self.handler_function = handler_function
        self.logger = logger
        self.threads = []

    def start(self):
        """Starts the threads"""
        if self.logger:
            self.logger.info("Starting %d threads", self.num_threads)
        self.threads = []
        for i in range(self.num_threads):
            thread = threading.Thread(target = self._thread)
            thread.setDaemon(1)
            thread.start()
            self.threads.append(thread)

    def _thread(self):
        """Takes tasks repeatedly the queue and calls the handler_function
        This is an internal method.
        """
        while 1:
            item = self.queue.get()
            if item is None:
                break
            try:
                self.handler_function(item)
            except:
                if self.logger:
                    self.logger.exception("Exception raised by"
                                          " threadpool handler function")
                else:
                    traceback.print_exc()

    def stop(self):
        """Stops the worker threads"""
        if self.logger:
            self.logger.info("Stopping threads")
        for i in range(len(self.threads)):
            self.put(None)
        for thread in self.threads:
            thread.join()

    def put(self, item):
        """Puts item into the queue to be processed by a worker thread"""
        self.queue.put(item)
