# Copyright (C) 2007, 2008 by Intevation GmbH
# Authors:
# Bernhard Herzog <bh@intevation.de>
#
# This program is free software under the GPL (>=v2)
# Read the file GPL.txt coming with the software for details.

import logging
import base64

from osaas.http.httpserver import HTTPRequestHandler
from osaas.http.threadedserver import ThreadedHTTPServer
from osaas.http.threadpool import ThreadPool
from recordfilter import default_filter
from userdb import UserDB
from formparser import parse_formdata, FormParserException


def parse_basic_authentication(authorization):
    """Parse a HTTP Basic authentication header and return (username, password).
    If the header cannot be parsed a ValueError exception is raised.
    """
    if authorization.startswith("Basic "):
        credentials_b64 = authorization.split()[1]
        credentials = base64.b64decode(credentials_b64)
        fields = credentials.split(":", 1)
        if len(fields) == 2:
            return fields
        else:
            raise ValueError("Malformed Basic auth informaion")
    else:
        raise ValueError("Unsupported authorization scheme")


class OSAASRequestHandler(HTTPRequestHandler):

    responses = HTTPRequestHandler.responses.copy()
    responses.update({
        411: ('Length Required', 'Client must specify Content-Length.'),
        415: ('Unsupported Media Type', 'Entity body in unsupported format.'),
        })

    def do_POST(self):
        try:
            self.handle_owsaccounting_formdata()
        except:
            self.log_exception("An unexpected exception occurred;"
                               " sending 500 response")
            self.send_error(500)

    def handle_owsaccounting_formdata(self):
        if self.path != "/owsaccounting/":
            self.send_error(404)
            return

        # check authentication
        if not self.check_user():
            return

        # check for the expected content type
        content_type = self.headers.getheader("Content-Type")
        if content_type != "application/x-www-form-urlencoded":
            self.send_error(415)
            return

        # If any data is associated with the request read it
        length = int(self.headers.getheader("Content-Length", "0"))
        if length:
            # FIXME: check whether length bytes were read
            data = self.rfile.read(length)
        else:
            self.send_error(411)
            return

        try:
            record = parse_formdata(data)
        except FormParserException, err:
            self.send_error(400, "Bad request: %s" % err)
            return

        if not default_filter(record):
            self.log_debug("record was filtered out: %s", record.raw_record)
            self.send_response(200)
            return

        try:
            record.validate_fields()
        except FormParserException, err:
            self.send_error(400, "Bad request: %s" % err)
            return

        self.log_debug("passing record to db thread")
        self.server.dbthreadpool.put(record)

        self.send_response(200)

    def check_user(self):
        authorization = self.headers.getheader("Authorization")
        try:
            if authorization:
                username, password = parse_basic_authentication(authorization)
                if self.server.userdb.check_credentials(username, password):
                    return True
        except:
            self.log_exception("Error parsing Authorization header")
        self.send_response(401)
        self.send_header("WWW-Authenticate", 'Basic realm="osaas"')
        return False

    def log_debug(self, *args, **kw):
        logger = logging.getLogger(self.server.error_logger_name)
        logger.debug(*args, **kw)

    def log_exception(self, *args, **kw):
        logger = logging.getLogger(self.server.error_logger_name)
        logger.debug(exc_info=1, *args, **kw)


class OSAASServer(ThreadedHTTPServer):

    def __init__(self, serveraddress, RequestHandlerClass=OSAASRequestHandler,
                 dbbackend=None, userdbfile=None,
                 **kw):
        ThreadedHTTPServer.__init__(self, serveraddress, RequestHandlerClass,
                                    **kw)
        self.userdb = UserDB()
        if userdbfile:
            self.userdb.read(userdbfile)
        logger = logging.getLogger(self.error_logger_name + ".dbpool")
        self.dbbackend = dbbackend
        if dbbackend is None:
            dbhandler = lambda r: None
        else:
            dbhandler = self.dbbackend.handle_record
        self.dbthreadpool = ThreadPool(1, dbhandler, logger=logger)
        self.dbthreadpool.start()

    def server_close(self):
        ThreadedHTTPServer.server_close(self)
        self.dbthreadpool.stop()
